@@ -1,4 +1,4 @@
-# $Id: common.rb 577 2010-04-11 15:39:20Z helixblue $
+# $Id: common.rb 615 2010-08-08 06:26:19Z helixblue $
 require 'fileutils'
 
 module Common
@@ -49,8 +49,13 @@ module Common
   ## finds a place to put temp files on the system ###################################
   def selectDirectory(dirs)
     dirs.compact.each do |dir|
-      if File.exists?(dir) && File.stat(dir).directory? && File.stat(dir).writable? 
-        return dir.gsub(/\\/, '/')
+      dir.gsub!(/\\/, '/')
+      debug "Checking #{dir} for writability."
+      if File.exists?(dir) && File.stat(dir).directory?
+        # write tests seem to be broken in Windows occassionaly.
+        if dir =~ /^\w:/ or File.stat(dir).writable?
+          return dir
+        end
       end
     end
     return Dir.pwd.gsub(/\\/, '/')
